@@ -5,7 +5,7 @@ import json
 import time
 from contextlib import contextmanager
 from threading import Thread
-from typing import Any, Callable, Iterator, NamedTuple, Optional
+from typing import Any, Callable, Dict, Iterator, NamedTuple, Optional
 
 import pika
 from pika.connection import Parameters
@@ -16,8 +16,12 @@ LOGGER = get_logger('rabbitmq_alphamoon')
 
 
 class IQueueConnector(metaclass=abc.ABCMeta):
+    @contextmanager
+    def open_connection(self) -> Iterator[RabbitMQConnector]:
+        pass
+
     @abc.abstractmethod
-    def consume_forever(self, process_message_callback: Callable[[Any], None]):
+    def consume_forever(self, process_message_callback: Callable[[Any, Dict[str, Any]], None]):
         pass
 
     @abc.abstractmethod
@@ -39,6 +43,8 @@ _RabbitMQConnection = NamedTuple(
     ],
 )
 
+_DEFAULT = None
+
 
 class RabbitMQConnector(IQueueConnector):
     def __init__(
@@ -48,12 +54,46 @@ class RabbitMQConnector(IQueueConnector):
         durable: bool = True,
         exclusive: bool = False,
         auto_delete: bool = False,
+        dead_letter_exchange: bool = False,
+        requeue_on_consumption_error=_DEFAULT,
     ):
+        """
+        Set up RabbitMQ connector.
+
+        Important note on `dead_letter_exchange` and `requeue_on_consumption_error`:
+
+            `dead_letter_exchange` causes `dlx_retry_count` incrementation for failed messages.
+
+            `requeue_on_consumption_error` causes messages for which processing failed
+            (i.e. message deserialization or processing callback failed) to be requeued again.
+            `redelivered` flag will be set to `True` for such messages.
+
+            Using both `dead_letter_exchange` and `requeue_on_consumption_error` may be undesired:
+                - rejected messages (processing failed) -> requeued
+                  (`redelivered` set to `True`, `dlx_retry_count` not incremented),
+                - expired messages -> DLX -> published again
+                  (`dlx_retry_count` incremented, `redelivered` set to `False`).
+
+        :param connection_parameters: connection parameters
+        :param queue: queue name
+        :param durable: whether to store queue on the disk (massages will remain after restarts)
+        :param exclusive:
+        :param auto_delete: automatically delete queue when disconnected
+        :param dead_letter_exchange: whether to set up DLX for failed messages
+        :param requeue_on_consumption_error: whether to retry messages if processing failed
+        """
         self._connection_parameters = connection_parameters
         self._queue = queue
         self._durable = durable
         self._exclusive = exclusive
         self._auto_delete = auto_delete
+        self._dead_letter_exchange = f'dlx.{self._queue}' if dead_letter_exchange else None
+
+        self._requeue = (  # when using DLX, don't requeue failed messages by default
+            requeue_on_consumption_error
+            if requeue_on_consumption_error is not _DEFAULT
+            else not dead_letter_exchange
+        )
 
         self._connection: Optional[_RabbitMQConnection] = None
         self._queue_declaration = None
@@ -63,12 +103,35 @@ class RabbitMQConnector(IQueueConnector):
         connection = pika.BlockingConnection(self._connection_parameters)
         channel = connection.channel()
         channel.confirm_delivery()
+
+        queue_arguments = {}
+        if self._dead_letter_exchange:
+            queue_arguments['x-dead-letter-exchange'] = self._dead_letter_exchange
+
         queue_declaration = channel.queue_declare(
             queue=self._queue,
             durable=self._durable,
             exclusive=self._exclusive,
             auto_delete=self._auto_delete,
+            arguments=queue_arguments,
         )
+
+        if self._dead_letter_exchange:
+            channel.exchange_declare(
+                exchange=self._dead_letter_exchange,
+                durable=self._durable,
+            )
+            channel.queue_bind(
+                exchange=self._dead_letter_exchange,
+                routing_key=self._queue,
+                queue=self._queue,
+            )
+            LOGGER.info(
+                "Dead letter exchange enabled, exchange %s, queue %s",
+                self._dead_letter_exchange,
+                self._queue,
+            )
+
         self._connection = _RabbitMQConnection(
             connection=connection,
             channel=channel,
@@ -118,7 +181,7 @@ class RabbitMQConnector(IQueueConnector):
         self._assure_connection()
         return self._get_opened_connection()
 
-    def consume_forever(self, process_message_callback: Callable[[Any], None]):
+    def consume_forever(self, process_message_callback: Callable[[Any, Dict[str, Any]], None]):
         connection = self._get_opened_connection()
 
         def callback_fun(
@@ -134,12 +197,71 @@ class RabbitMQConnector(IQueueConnector):
                 channel.basic_ack(delivery_tag=tag)
                 LOGGER.info("Sent ack for message # %s, queue %s", tag, self._queue)
 
-            def process_fun():
-                LOGGER.debug("Deserializing body of message # %s, queue %s", tag, self._queue)
-                message = json.loads(body)
+            def send_reject():
+                channel.basic_reject(delivery_tag=tag, requeue=self._requeue)
+                LOGGER.info(
+                    "Sent reject with requeue=%s for message # %s, queue %s",
+                    self._requeue,
+                    tag,
+                    self._queue,
+                )
 
-                LOGGER.info("Processing message # %s, queue %s", tag, self._queue)
-                process_message_callback(message)
+            def process_fun():
+                try:
+                    LOGGER.debug("Deserializing body of message # %s, queue %s", tag, self._queue)
+                    message = json.loads(body)
+                except Exception as error:
+                    LOGGER.error(
+                        "Deserializing body failed for message of # %s, queue %s, %s: %s",
+                        tag,
+                        self._queue,
+                        type(error).__name__,
+                        error,
+                    )
+                    try:
+                        LOGGER.debug("Sending reject for message # %s, queue %s", tag, self._queue)
+                        channel.connection.add_callback_threadsafe(send_reject)
+                    except pika.exceptions.AMQPError as amqp_error:
+                        LOGGER.error(
+                            "Sending reject failed for message # %s, queue %s, %s: %s",
+                            tag,
+                            self._queue,
+                            type(amqp_error).__name__,
+                            amqp_error,
+                        )
+                    raise
+
+                info = {}
+                if self._requeue:
+                    info['redelivered'] = deliver.redelivered
+                if self._dead_letter_exchange:
+                    x_death_header = (properties.headers or {}).get('x-death', [{}])[0]
+                    info['dlx_retry_count'] = x_death_header.get('count', 0)
+                    info['dlx_last_fail_reason'] = x_death_header.get('reason')
+
+                try:
+                    LOGGER.info("Processing message # %s, queue %s", tag, self._queue)
+                    process_message_callback(message, info)
+                except Exception as error:
+                    LOGGER.error(
+                        "Processing message in callback failed for # %s, queue %s, %s: %s",
+                        tag,
+                        self._queue,
+                        type(error).__name__,
+                        error,
+                    )
+                    try:
+                        LOGGER.debug("Sending reject for message # %s, queue %s", tag, self._queue)
+                        channel.connection.add_callback_threadsafe(send_reject)
+                    except pika.exceptions.AMQPError as amqp_error:
+                        LOGGER.error(
+                            "Sending reject failed for message # %s, queue %s, %s: %s",
+                            tag,
+                            self._queue,
+                            type(amqp_error).__name__,
+                            amqp_error,
+                        )
+                    raise
 
                 try:
                     LOGGER.debug("Sending ack for message # %s, queue %s", tag, self._queue)
